﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;

namespace PojeClient {
    // Stateが2つ含むことになる，盤面の状態．
    class Board {
        public const int WALL = -2;
        public const int JAMSTONE = -1;

        // 意思決定のための本質的なもの
        private int[,] fStones;            // 盤上の石の色．座標は１から始まる．-2は壁，-1は邪魔珠，0はなにもなし，1-4が各色の珠．
        private JamInfo fJamInfo;          // 現在の攻撃されてる珠数

        private int[] fNextColors;         // 操作する石の色．0,1 が現在のもの，2-5がnextのもの
        private List<int> fLastColors;     // これまでに来た石の色．
        private int fChainCount;           // 現在連鎖中かどうか．0なら連鎖中でない．n>=1 ならn連鎖めが終わったところ（であって，次は操作できない）
        

        // 一時的なもの，他から計算できるもの
        private List<Point> fVanishPoints; // 消える箇所（邪魔珠含む）
        private int fVanishColorStonesCount;// 消える珠数（邪魔珠含まない）
        private int[] fHeights;            // 各X座標の位置の，珠の高さ．座標は1から始まる．この値は 0（珠なし）から BoardHeight（全て埋まってる）まである．

        private List<Action> fPossibleActions;   // 可能な着手集合
        private Action fShownAction;           // 現在表示中の着手（必ずしも fPossibleActionsに含まれている必要はない）








        // 盤面を初期化する
        public void initialize(int[] nextColors) {
            fJamInfo = new JamInfo();
            fLastColors = new List<int>();
            fNextColors = new int[nextColors.Length];
            for (int i = 0; i < nextColors.Length; i++) {
                fNextColors[i] = nextColors[i];
                fLastColors.Add(nextColors[i]);
            }
            fChainCount = 0;

            fStones = new int[Settings.BoardWidth + 2, Settings.BoardHeight + 2];
            fHeights = new int[Settings.BoardWidth + 2];
            // 7x11 の範囲の外側を，壁で埋める
            for (int x=0; x<=Settings.BoardWidth+1; x++) for (int y=0; y<=Settings.BoardHeight+1;y++) fStones[x,y]=WALL;
            // 7x11 の範囲を空にする
            for (int x=1; x<=Settings.BoardWidth; x++) {
                fHeights[x] = 0;
                for (int y=1; y<=Settings.BoardHeight; y++) fStones[x,y]=0;
            }
        }

        // デフォルトコンストラクタ．initialize とセットで使う
        public Board() {
        }

        // コピー用のコンストラクタ．完全なコピーではない．特に fPossibleActions がコピーされないことに注意
        public Board(Board src) {
            fStones = new int[Settings.BoardWidth + 2, Settings.BoardHeight + 2];
            fHeights = new int[Settings.BoardWidth + 2];
            for (int x = 0; x <= Settings.BoardWidth + 1; x++) {
                fHeights[x] = src.getHeight(x);
                for (int y = 0; y <= Settings.BoardHeight + 1; y++) fStones[x, y] = src.getStone(x, y);
            }
            fJamInfo = new JamInfo(src.getJamInfo());

            int[] nexts = src.getNextColors();
            fNextColors = new int[nexts.Length];
            for (int i = 0; i < nexts.Length; i++) fNextColors[i] = nexts[i];

            int[] lasts = src.getLastColors(int.MaxValue);
            fLastColors = new List<int>();
            for (int i = 0; i < lasts.Length; i++) fLastColors.Add(lasts[i]);

            fChainCount = src.getChainCount();
        }
            


        #region //ゲッタ セッタ------------------------

        public int getChainCount() {
            return fChainCount;
        }

        public void setChainCount(int count) {
            fChainCount = count;
        }
            

        public int getVanishColorStonesCount() {
            return fVanishColorStonesCount;
        }

        public int getStone(int x, int y) {
            return fStones[x,y];
        }


        public int getHeight(int x) {
            return fHeights[x];
        }

        // 埋まった＝ fHeightが全部 BoardHeightか
        public bool isFilled() {
            for (int x = 1; x <= Settings.BoardWidth; x++) {
                if (fHeights[x] < Settings.BoardHeight) return false;
            }
            return true;
        }

        public JamInfo getJamInfo() {
            return fJamInfo;
        }

        public List<Point> getVanishPoints() {
            return fVanishPoints;
        }



        public Action getShownAction() {
            return fShownAction;
        }


        public void setShownAction(Action a) {
            fShownAction = a;
        }

        public List<Action> getPossibleActions() {
            return fPossibleActions;
        }
        
        // fLastColors のうち最後のcount個を返す
        public int[] getLastColors(int count) {
            if (count > fLastColors.Count) {
                count = fLastColors.Count;
            }
            int[] colors = new int[count];
            for (int i = 0; i < count; i++) {
                colors[i] = fLastColors[fLastColors.Count - count + i];
            }
            return colors;
        }

        public int[] getNextColors() {
            return fNextColors;
        }

        #endregion

        // NextColors をずらして，新しい色を入れる
        public void pushNextColors(int[] colors) {
            for (int i = 0; i < 2 * Settings.NextOpenTurns - 2; i++) {
                fNextColors[i] = fNextColors[i + 2];
            }
            fNextColors[2 * Settings.NextOpenTurns - 2] = colors[0];
            fNextColors[2 * Settings.NextOpenTurns - 1] = colors[1];
            fLastColors.Add(colors[0]);
            fLastColors.Add(colors[1]);
        }

        // 普通のセッタ
        public void setNextColor(int color, int index) {
            fNextColors[index] = color;
        }



        // 行動したあとに，fNextColors[0],[1] に色が残っているのをとりあえず消す
        public void deleteCurrentColors() {
            fNextColors[0] = 0;
            fNextColors[1] = 0;
        }




        //------　少し重い処理たち


        // 可能なActionの集合を作る
        // ver802 色珠が１個だけの場合がある
        public List<Action> createPossibleActions() {
            fPossibleActions = new List<Action>();
            for (int x = 1; x <= Settings.BoardWidth; x++) {
                int maxR = 3;
                if (fNextColors[0] == fNextColors[1]) maxR = 1; // 色同じときは縦横２パターンしかない
                if (fNextColors[1] == 0) maxR = 0;              // １個だけの場合，回転には意味がない
                for (int r = 0; r <= maxR; r++) {
                    // はみ出す場合はダメ
                    if (x + Settings.RotateX[r] == 0 || x + Settings.RotateX[r] == Settings.BoardWidth + 1) continue;

                    // 両方とも最上部まで埋まっている場合はダメ
                    Action a = new Action(x, r, this, fNextColors[0], fNextColors[1]);
                    if (a.isLegal() == false) continue;

                    fPossibleActions.Add(a);
                }
            }
            return fPossibleActions;
        }

        // 状態遷移０ステップ目，邪魔珠が落ちる
        public void transit0_JamDrop(int count) {
            int x = fJamInfo.getFallIndex(); // 次にどこに落ちるか
            for (int i = 0; i < count; i++) {
                if (fHeights[x] < Settings.BoardHeight) {  // 上まで詰まっていたらもう降らない
                    fHeights[x]++;
                    fStones[x, fHeights[x]] = JAMSTONE;
                }
                x = (x % Settings.BoardWidth) + 1;  // 次に落ちるのはその右側
            }
            fJamInfo.setFallIndex(x);
        }


        // 状態遷移１ステップ目，珠を落とす（盤上に追加する）． fHeightの加算もする．落ちた地点を返す．
        // ver802 pointsが 1個だけの場合がある
        public List<Point> transit1_ActionDrop(Action a) {
            List<Point> drops = new List<Point>();
            Point[] points = a.getDropPoints();
            for (int i = 0; i < points.Length; i++) {
                int x = points[i].X;
                int y = points[i].Y;
                if (y <= Settings.BoardHeight) {
                    fStones[x, y] = a.getColor(i);
                    fHeights[x]++;
                    drops.Add(new Point(x, y));
                }
            }
            return drops;
        }

        // 状態遷移１ステップ目，連鎖中の場合は浮いてる珠を落とす. fHeightの計算もしなおす．落ちた地点を返す．
        public List<Point> transit1_ChainDrop() {
            List<Point> drops = new List<Point>();
            for (int x = 1; x <= Settings.BoardWidth; x++) {
                for (int ybase = 1; ybase <= fHeights[x] - 1; ybase++) {
                    // この場所が空で，その上に珠があるなら持ってくる
                    if (fStones[x, ybase] != 0) continue;
                    int existY = -1;
                    for (int y = ybase + 1; y <= fHeights[x]; y++) {
                        if (fStones[x, y] > 0 || fStones[x, y] == JAMSTONE) {
                            existY = y;
                            break;
                        }
                    }
                    if (existY == -1) {
                        // existY=-1 のままなら，これ以上落ちる石はないので終了
                        break;
                    } else {
                        // 見つけた珠を消して今の場所に置く
                        fStones[x, ybase] = fStones[x, existY];
                        fStones[x, existY] = 0;
                        // Color石の場合，落ちたリストに含める
                        if (fStones[x, ybase] > 0) drops.Add(new Point(x, ybase));
                    }
                }

                // heightの計算しなおし，ちょっと効率悪いけど．
                fHeights[x] = 0;
                while (fStones[x, fHeights[x] + 1] > 0 || fStones[x, fHeights[x] + 1] == JAMSTONE) fHeights[x]++;
            }
            return drops;
        }


        // 状態遷移２ステップ目
        // 珠が落ちた地点周囲を見て，どの点が消えるかを計算する．
        // fVanishPoints リストを作成，fVanishColorStonesCountを計算，fVanishColorStonesCount 自体も返す
        public int transit2_calcVanishPoints(List<Point> dropStones) {

            int wd = Settings.BoardWidth;
            int hg = Settings.BoardHeight;
            int[,] groupLength = new int[wd + 2, hg + 2];
            for (int x = 0; x <= wd + 1; x++) for (int y = 0; y <= hg + 1; y++) groupLength[x, y] = 0;

            // まず，注目している地点を含むグループの個数を計算
            foreach (Point p in dropStones) {
                if (groupLength[p.X, p.Y] > 0 || fStones[p.X, p.Y] <= 0) continue;   //すでにカウント済みか，空所・邪魔珠の場合は抜ける

                // その珠を含むグル―プを生成
                List<Point> points = new List<Point>();
                points.Add(new Point(p.X, p.Y));
                groupLength[p.X, p.Y] = -1; //グループに含めた一時フラグ
                int lastCount = 0; // この個数は処理した  
                int col = fStones[p.X, p.Y];
                // 数が増えなくなるまで繰り返す
                while (lastCount < points.Count) {
                    // 最後に追加された箇所の上下左右を見る
                    int toCount = points.Count;
                    for (int i = lastCount; i < toCount; i++) {
                        Point neighbor = points[i];
                        for (int r = 0; r <= 3; r++) {
                            int nx = neighbor.X + Settings.RotateX[r];
                            int ny = neighbor.Y + Settings.RotateY[r];
                            if (groupLength[nx, ny] == 0 && fStones[nx, ny] == col) {
                                points.Add(new Point(nx, ny));
                                groupLength[nx, ny] = -1;
                            }
                        }
                    }
                    lastCount = toCount;
                }

                // そのグループのメンバーの GroupLengthを更新
                foreach (Point member in points) {
                    groupLength[member.X, member.Y] = points.Count;
                }
            }

            // 消える地点をカウント
            fVanishColorStonesCount =0;
            fVanishPoints = new List<Point>();
            for (int x=1; x<=Settings.BoardWidth; x++) {
                for (int y=1; y<=Settings.BoardHeight; y++) {
                    if (fStones[x,y]>0) {
                        // 色珠の場合，４個以上繋がっていれば消える
                        if (groupLength[x, y] >= Settings.NumVanish) {
                            fVanishPoints.Add(new Point(x, y));
                            fVanishColorStonesCount++;
                        }
                    } else if (fStones[x,y]==JAMSTONE) {
                        // 邪魔珠の場合，上下左右に消える石があれば消える
                        for (int r = 0; r <= 3; r++) {
                            if (groupLength[x + Settings.RotateX[r], y + Settings.RotateY[r]] >= Settings.NumVanish) {
                                fVanishPoints.Add(new Point(x, y));
                                break;
                            }
                        }
                    }
                }
            }
            return fVanishColorStonesCount;
        }

        // 状態遷移３ステップ目
        // fVanishPoints の部分を消す（落下はあとなのでfHeightsの計算もしない）
        public void transit3_Vanish() {
            foreach (Point p in fVanishPoints) {
                fStones[p.X,p.Y] = 0;
            }
        }




        // デバッグ用．fHeight と fStone の現状を見る
        public void checkBoard() {
            String s = "";
            for (int x = 1; x <= Settings.BoardWidth; x++) s = s + fHeights[x] + " ";
            for (int y = Settings.BoardHeight; y >= 1; y--) {
                s = s + "\n";
                for (int x = 1; x <= Settings.BoardWidth; x++) s = s + fStones[x, y] + " ";
            }
            MessageBox.Show(s);
        }


    }
}
